# Copyright (C) 2020-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import collections
from contextlib import contextmanager
import hashlib
import math
from pathlib import Path
import tempfile

import pyorc
import pytest

from swh.export.exporters import orc
from swh.export.relational import MAIN_TABLES, RELATION_TABLES
from swh.model.model import (
    Content,
    Directory,
    ModelObjectType,
    Origin,
    OriginVisit,
    OriginVisitStatus,
    Release,
    Revision,
    SkippedContent,
    Snapshot,
    TimestampWithTimezone,
)
from swh.model.tests.swh_model_data import TEST_OBJECTS
from swh.objstorage.factory import get_objstorage


@contextmanager
def orc_tmpdir(tmpdir):
    if tmpdir:
        yield Path(tmpdir)
    else:
        with tempfile.TemporaryDirectory() as tmpdir:
            yield Path(tmpdir)


@contextmanager
def orc_sensitive_tmpdir(tmpdir):
    if tmpdir:
        yield Path(tmpdir)
    else:
        with tempfile.TemporaryDirectory() as tmpdir:
            yield Path(tmpdir)


@contextmanager
def orc_export(messages, config=None, tmpdir=None, sensitive_tmpdir=None):
    with orc_tmpdir(tmpdir) as tmpdir:
        with orc_sensitive_tmpdir(sensitive_tmpdir) as sensitive_tmpdir:
            if config is None:
                config = {}
            with orc.ORCExporter(config, tmpdir, sensitive_tmpdir) as exporter:
                for object_type, objects in messages.items():
                    for obj in objects:
                        exporter.process_object(object_type, obj)
            yield tmpdir


def orc_load(rootdir):
    res = collections.defaultdict(list)
    res["rootdir"] = rootdir
    for obj_type_dir in rootdir.iterdir():
        for orc_file in obj_type_dir.iterdir():
            with orc_file.open("rb") as orc_obj:
                reader = pyorc.Reader(
                    orc_obj,
                    converters={pyorc.TypeKind.TIMESTAMP: orc.SWHTimestampConverter},
                )
                obj_type = reader.user_metadata["swh_object_type"].decode()
                res[obj_type].extend(reader)
    return res


def exporter(messages, config=None, tmpdir=None):
    with orc_export(messages, config, tmpdir) as exportdir:
        return orc_load(exportdir)


def test_export_origin():
    obj_type = Origin.object_type
    output = exporter({obj_type: TEST_OBJECTS[obj_type]})
    for obj in TEST_OBJECTS[obj_type]:
        sha1 = hashlib.sha1(obj.url.encode()).hexdigest()
        assert (sha1, obj.url) in output[obj_type.value]


def test_export_origin_visit():
    obj_type = OriginVisit.object_type
    output = exporter({obj_type: TEST_OBJECTS[obj_type]})
    for obj in TEST_OBJECTS[obj_type]:
        assert (
            obj.origin,
            obj.visit,
            orc.datetime_to_tuple(obj.date),
            obj.type,
        ) in output[obj_type.value]


def test_export_origin_visit_status():
    obj_type = OriginVisitStatus.object_type
    output = exporter({obj_type: TEST_OBJECTS[obj_type]})
    for obj in TEST_OBJECTS[obj_type]:
        assert (
            obj.origin,
            obj.visit,
            orc.datetime_to_tuple(obj.date),
            obj.status,
            orc.hash_to_hex_or_none(obj.snapshot),
            obj.type,
        ) in output[obj_type.value]


def test_export_snapshot():
    obj_type = Snapshot.object_type
    output = exporter({obj_type: TEST_OBJECTS[obj_type]})
    for obj in TEST_OBJECTS[obj_type]:
        assert (orc.hash_to_hex_or_none(obj.id),) in output["snapshot"]
        for branch_name, branch in obj.branches.items():
            if branch is None:
                continue
            assert (
                orc.hash_to_hex_or_none(obj.id),
                branch_name,
                orc.hash_to_hex_or_none(branch.target),
                str(branch.target_type.value),
            ) in output["snapshot_branch"]


def test_export_release():
    obj_type = Release.object_type
    output = exporter({obj_type: TEST_OBJECTS[obj_type]})
    for obj in TEST_OBJECTS[obj_type]:
        assert (
            orc.hash_to_hex_or_none(obj.id),
            obj.name,
            obj.message,
            orc.hash_to_hex_or_none(obj.target),
            obj.target_type.value,
            obj.author.fullname if obj.author else None,
            *orc.swh_date_to_tuple(getattr(obj, "date", None)),
            obj.raw_manifest,
        ) in output[obj_type.value]


def test_export_revision():
    obj_type = Revision.object_type
    output = exporter({obj_type: TEST_OBJECTS[obj_type]})
    for obj in TEST_OBJECTS[obj_type]:
        assert (
            orc.hash_to_hex_or_none(obj.id),
            obj.message,
            obj.author.fullname,
            *orc.swh_date_to_tuple(getattr(obj, "date", None)),
            obj.committer.fullname,
            *orc.swh_date_to_tuple(getattr(obj, "committer_date", None)),
            orc.hash_to_hex_or_none(obj.directory),
            obj.type.value,
            obj.raw_manifest,
        ) in output["revision"]
        for i, parent in enumerate(obj.parents):
            assert (
                orc.hash_to_hex_or_none(obj.id),
                orc.hash_to_hex_or_none(parent),
                i,
            ) in output["revision_history"]


def test_export_directory():
    obj_type = Directory.object_type
    output = exporter({obj_type: TEST_OBJECTS[obj_type]})
    for obj in TEST_OBJECTS[obj_type]:
        assert (orc.hash_to_hex_or_none(obj.id), obj.raw_manifest) in output[
            "directory"
        ]
        for entry in obj.entries:
            assert (
                orc.hash_to_hex_or_none(obj.id),
                entry.name,
                entry.type,
                orc.hash_to_hex_or_none(entry.target),
                entry.perms,
            ) in output["directory_entry"]


def test_export_content():
    obj_type = Content.object_type
    output = exporter({obj_type: TEST_OBJECTS[obj_type]})
    for obj in TEST_OBJECTS[obj_type]:
        assert (
            orc.hash_to_hex_or_none(obj.sha1),
            orc.hash_to_hex_or_none(obj.sha1_git),
            orc.hash_to_hex_or_none(obj.sha256),
            orc.hash_to_hex_or_none(obj.blake2s256),
            obj.length,
            obj.status,
            None,
        ) in output[obj_type.value]


def test_export_skipped_content():
    obj_type = SkippedContent.object_type
    output = exporter({obj_type: TEST_OBJECTS[obj_type]})
    for obj in TEST_OBJECTS[obj_type]:
        assert (
            orc.hash_to_hex_or_none(obj.sha1),
            orc.hash_to_hex_or_none(obj.sha1_git),
            orc.hash_to_hex_or_none(obj.sha256),
            orc.hash_to_hex_or_none(obj.blake2s256),
            obj.length,
            obj.status,
            obj.reason,
        ) in output[obj_type.value]


def test_date_to_tuple():
    ts = {"seconds": 123456, "microseconds": 1515}
    assert orc.swh_date_to_tuple(
        TimestampWithTimezone.from_dict({"timestamp": ts, "offset_bytes": b"+0100"})
    ) == (
        (123456, 1515),
        60,
        b"+0100",
    )

    assert orc.swh_date_to_tuple(
        TimestampWithTimezone.from_dict(
            {
                "timestamp": ts,
                "offset": 120,
                "negative_utc": False,
                "offset_bytes": b"+0100",
            }
        )
    ) == ((123456, 1515), 60, b"+0100")

    assert orc.swh_date_to_tuple(
        TimestampWithTimezone.from_dict(
            {
                "timestamp": ts,
                "offset": 120,
                "negative_utc": False,
            }
        )
    ) == ((123456, 1515), 120, b"+0200")

    assert orc.swh_date_to_tuple(
        TimestampWithTimezone.from_dict(
            {
                "timestamp": ts,
                "offset": 0,
                "negative_utc": True,
            }
        )
    ) == (
        (123456, 1515),
        0,
        b"-0000",
    )


# mapping of related tables for each main table (if any)
RELATED = {
    "snapshot": ["snapshot_branch"],
    "revision": ["revision_history", "revision_extra_headers"],
    "directory": ["directory_entry"],
}


@pytest.mark.parametrize(
    "obj_type",
    (ModelObjectType(obj_type) for obj_type in MAIN_TABLES.keys()),
)
@pytest.mark.parametrize("max_rows", (None, 1, 2, 10000))
def test_export_related_files(max_rows, obj_type, tmpdir):
    config = {"orc": {}}
    if max_rows is not None:
        config["orc"]["max_rows"] = {obj_type.value: max_rows}
    exporter(
        {obj_type: TEST_OBJECTS[obj_type]},
        config=config,
        tmpdir=tmpdir,
    )
    # check there are as many ORC files as objects
    orcfiles = [
        fname for fname in (tmpdir / obj_type.value).listdir(f"{obj_type}-*.orc")
    ]
    if max_rows is None:
        assert len(orcfiles) == 1
    else:
        assert len(orcfiles) == math.ceil(len(TEST_OBJECTS[obj_type]) / max_rows)
    # check the number of related ORC files
    for related in RELATED.get(obj_type.value, ()):
        related_orcfiles = [
            fname for fname in (tmpdir / related).listdir(f"{related}-*.orc")
        ]
        assert len(related_orcfiles) == len(orcfiles)

    # for each ORC file, check related files only reference objects in the
    # corresponding main table
    for orc_file in orcfiles:
        with orc_file.open("rb") as orc_obj:
            reader = pyorc.Reader(
                orc_obj,
                converters={pyorc.TypeKind.TIMESTAMP: orc.SWHTimestampConverter},
            )
            uuid = reader.user_metadata["swh_uuid"].decode()
            assert orc_file.basename == f"{obj_type}-{uuid}.orc"
            rows = list(reader)
            obj_ids = [row[0] for row in rows]

        # check the related tables
        for related in RELATED.get(obj_type.value, ()):
            orc_file = tmpdir / related / f"{related}-{uuid}.orc"
            with orc_file.open("rb") as orc_obj:
                reader = pyorc.Reader(
                    orc_obj,
                    converters={pyorc.TypeKind.TIMESTAMP: orc.SWHTimestampConverter},
                )
                assert reader.user_metadata["swh_uuid"].decode() == uuid
                rows = list(reader)
                # check branches in this file only concern current snapshot (obj_id)
                for row in rows:
                    assert row[0] in obj_ids


@pytest.mark.parametrize(
    "obj_type",
    (ModelObjectType(obj_type) for obj_type in MAIN_TABLES.keys()),
)
def test_export_related_files_separated(obj_type, tmpdir):
    exporter({obj_type: TEST_OBJECTS[obj_type]}, tmpdir=tmpdir)
    # check there are as many ORC files as objects
    orcfiles = [
        fname for fname in (tmpdir / obj_type.value).listdir(f"{obj_type}-*.orc")
    ]
    assert len(orcfiles) == 1
    # check related ORC files are in their own directory
    for related in RELATED.get(obj_type.value, ()):
        related_orcfiles = [
            fname for fname in (tmpdir / related).listdir(f"{related}-*.orc")
        ]
        assert len(related_orcfiles) == len(orcfiles)


@pytest.mark.parametrize("table_name", RELATION_TABLES.keys())
def test_export_invalid_max_rows(table_name):
    config = {"orc": {"max_rows": {table_name: 10}}}
    with pytest.raises(ValueError):
        exporter({}, config=config)


def test_export_content_with_data(monkeypatch, tmpdir):
    obj_type = Content.object_type
    objstorage = get_objstorage("memory")
    for content in TEST_OBJECTS[obj_type]:
        objstorage.add(content=content.data, obj_id=content.hashes())

    def get_objstorage_mock(**kw):
        if kw.get("cls") == "mock":
            return objstorage

    monkeypatch.setattr(orc, "get_objstorage", get_objstorage_mock)
    config = {
        "orc": {
            "with_data": True,
            "objstorage": {"cls": "mock"},
        },
    }

    output = exporter(
        {obj_type: TEST_OBJECTS[obj_type]},
        config=config,
        tmpdir=tmpdir,
    )
    for obj in TEST_OBJECTS[obj_type]:
        assert (
            orc.hash_to_hex_or_none(obj.sha1),
            orc.hash_to_hex_or_none(obj.sha1_git),
            orc.hash_to_hex_or_none(obj.sha256),
            orc.hash_to_hex_or_none(obj.blake2s256),
            obj.length,
            obj.status,
            obj.data,
        ) in output[obj_type.value]



# Copyright (C) 2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
import os
import pathlib
import sys
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set

import click

from swh.core.cli import CONTEXT_SETTINGS
from swh.core.cli import swh as swh_cli_group

from .relational import MAIN_TABLES

if TYPE_CHECKING:
    from swh.model.swhids import ExtendedSWHID


@swh_cli_group.group(name="export", context_settings=CONTEXT_SETTINGS)
@click.option(
    "--config-file",
    "-C",
    default=None,
    type=click.Path(exists=True, dir_okay=False),
    help="Configuration file.",
)
@click.pass_context
def export_cli_group(ctx, config_file):
    """Dataset Export Tools.

    A set of tools to export datasets from the Software Heritage Archive in
    various formats.

    """
    from swh.core import config

    ctx.ensure_object(dict)
    conf = config.read(config_file)
    ctx.obj["config"] = conf


@export_cli_group.group("graph")
@click.pass_context
def graph(ctx):
    """Manage graph export"""
    pass


AVAILABLE_EXPORTERS = {
    "edges": "swh.export.exporters.edges:GraphEdgesExporter",
    "orc": "swh.export.exporters.orc:ORCExporter",
}


@graph.command("export")
@click.argument("export-path", type=click.Path())
@click.option(
    "--export-id",
    "-e",
    help=(
        "Unique ID of the export run. This is appended to the kafka "
        "group_id config file option. If group_id is not set in the "
        "'journal' section of the config file, defaults to 'swh-export-export-'."
    ),
)
@click.option(
    "--formats",
    "-f",
    type=click.STRING,
    default=",".join(AVAILABLE_EXPORTERS.keys()),
    show_default=True,
    help="Formats to export.",
)
@click.option("--processes", "-p", default=1, help="Number of parallel processes")
@click.option(
    "--exclude",
    type=click.STRING,
    help="Comma-separated list of object types to exclude",
)
@click.option(
    "--types",
    "object_types",
    type=click.STRING,
    help="Comma-separated list of objects types to export",
)
@click.option(
    "--margin",
    type=click.FloatRange(0, 1),
    help=(
        "Offset margin to start consuming from. E.g. is set to '0.95', "
        "consumers will start at 95% of the last committed offset; "
        "in other words, start earlier than last committed position."
    ),
)
@click.option(
    "--sensitive-export-path",
    type=click.STRING,
    help="Path where sensitive data (eg. fullnames) will be exported to.",
)
@click.pass_context
def export_graph(
    ctx, export_path, sensitive_export_path, formats, exclude, object_types, **kwargs
):
    """Export the Software Heritage graph as an edge dataset."""

    config = ctx.obj["config"]
    if object_types:
        object_types = {o.strip() for o in object_types.split(",")}
        invalid_object_types = object_types - set(MAIN_TABLES.keys())
        if invalid_object_types:
            raise click.BadOptionUsage(
                option_name="types",
                message=f"Invalid object types: {', '.join(invalid_object_types)}.",
            )
    else:
        object_types = set(MAIN_TABLES.keys())
    exclude_obj_types = {o.strip() for o in (exclude.split(",") if exclude else [])}
    export_formats = [c.strip() for c in formats.split(",")]
    for f in export_formats:
        if f not in AVAILABLE_EXPORTERS:
            raise click.BadOptionUsage(
                option_name="formats", message=f"{f} is not an available format."
            )
    export_path = pathlib.Path(export_path)
    if sensitive_export_path is not None:
        sensitive_export_path = pathlib.Path(sensitive_export_path)
    else:
        sensitive_export_path = export_path.parent / f"{export_path.name}-sensitive"

    run_export_graph(
        config,
        pathlib.Path(export_path),
        pathlib.Path(sensitive_export_path),
        export_formats,
        list(object_types),
        exclude_obj_types=exclude_obj_types,
        **kwargs,
    )


def get_masked_swhids(logger, config: Dict[str, Any]) -> Set["ExtendedSWHID"]:
    """Fetches the masking database and returns the list of all non-visible SWHIDs"""
    import tqdm

    from swh.storage.proxies.masking.db import MaskingQuery

    if config["masking_db"] is None:
        logger.warning("Exporting dataset without masking.")
        return set()
    masking_query = MaskingQuery.connect(config["masking_db"])
    return {
        swhid
        for (swhid, statuses) in tqdm.tqdm(
            masking_query.iter_masked_swhids(),
            desc="Listing masked SWHIDs",
            unit_scale=True,
        )
    }


def run_export_graph(
    config: Dict[str, Any],
    export_path: pathlib.Path,
    sensitive_export_path: pathlib.Path,
    export_formats: List[str],
    object_types: List[str],
    exclude_obj_types: Set[str],
    export_id: Optional[str],
    processes: int,
    margin: Optional[float],
):
    import functools
    from importlib import import_module
    import logging
    import resource
    import tempfile
    import uuid

    from .fullnames import process_fullnames
    from .journalprocessor import ParallelJournalProcessor

    logger = logging.getLogger(__name__)

    masked_swhids = get_masked_swhids(logger, config)

    if not export_id:
        export_id = str(uuid.uuid4())

    # Enforce order (from origin to contents) to reduce number of holes in the graph.
    object_types = [
        obj_type for obj_type in MAIN_TABLES.keys() if obj_type in object_types
    ]

    # ParallelJournalProcessor opens 256 LevelDBs in total. Depending on the number of
    # processes, this can exceed the maximum number of file descriptors (soft limit
    # defaults to 1024 on Debian), so let's increase it.
    (soft, hard) = resource.getrlimit(resource.RLIMIT_NOFILE)
    nb_shards = 256  # TODO: make this configurable or detect nb of kafka partitions
    open_fds_per_shard = 61  # estimated with plyvel==1.3.0 and libleveldb1d==1.22-3
    spare = 1024  # for everything other than LevelDB
    want_fd = nb_shards * open_fds_per_shard + spare
    if hard < want_fd:
        logger.warning(
            "Hard limit of open file descriptors (%d) is lower than ideal (%d)",
            hard,
            want_fd,
        )
    if soft < want_fd:
        want_fd = min(want_fd, hard)
        logger.info(
            "Soft limit of open file descriptors (%d) is too low, increasing to %d",
            soft,
            want_fd,
        )
        resource.setrlimit(resource.RLIMIT_NOFILE, (want_fd, hard))

    def importcls(clspath):
        mod, cls = clspath.split(":")
        m = import_module(mod)
        return getattr(m, cls)

    exporter_cls = dict(
        (fmt, importcls(clspath))
        for (fmt, clspath) in AVAILABLE_EXPORTERS.items()
        if fmt in export_formats
    )

    with tempfile.TemporaryDirectory() as csv_dir:
        dup_dir = pathlib.Path(csv_dir) / "duplicates"
        dup_dir.mkdir(parents=True, exist_ok=True)
        dedup_dir = pathlib.Path(csv_dir) / "deduplicated"
        dedup_dir.mkdir(parents=True, exist_ok=True)

        # Run the exporter for each edge type.
        parallel_exporters = {}
        for obj_type in object_types:
            if obj_type in exclude_obj_types:
                continue
            exporters = [
                functools.partial(
                    exporter_cls[f],
                    config=config,
                    export_path=export_path / f,
                    sensitive_export_path=sensitive_export_path / f,
                )
                for f in export_formats
            ]
            parallel_exporters[obj_type] = ParallelJournalProcessor(
                config,
                masked_swhids,
                exporters,
                export_id,
                obj_type,
                node_sets_path=export_path / ".node_sets",
                dup_dir=dup_dir,
                dedup_dir=dedup_dir,
                processes=processes,
                offset_margin=margin,
            )
            # Fetch all offsets before we start exporting to minimize the time interval
            # between the offsets of each topic
            parallel_exporters[obj_type].get_offsets()

        for obj_type, parallel_exporter in parallel_exporters.items():
            parallel_exporter.run()

        if config["journal"].get("privileged"):
            fullnames_export_path = sensitive_export_path / "orc" / "person"
            fullnames_export_path.mkdir(parents=True, exist_ok=True)
            fullnames_orc = fullnames_export_path / f"{uuid.uuid4()}.orc"
            process_fullnames(fullnames_orc, dedup_dir)


@graph.command("sort")
@click.argument("export-path", type=click.Path())
@click.pass_context
def sort_graph(ctx, export_path):
    config = ctx.obj["config"]
    from .exporters.edges import sort_graph_nodes

    sort_graph_nodes(export_path, config)


@export_cli_group.group("athena")
@click.pass_context
def athena(ctx):
    """Manage and query a remote AWS Athena database"""
    pass


@athena.command("create")
@click.option(
    "--database-name", "-d", default="swh", help="Name of the database to create"
)
@click.option(
    "--location-prefix",
    "-l",
    required=True,
    help="S3 prefix where the dataset can be found",
)
@click.option(
    "-o", "--output-location", help="S3 prefix where results should be stored"
)
@click.option(
    "-r", "--replace-tables", is_flag=True, help="Replace the tables that already exist"
)
def athena_create(
    database_name, location_prefix, output_location=None, replace_tables=False
):
    """Create tables on AWS Athena pointing to a given graph dataset on S3."""
    from .athena import create_tables

    create_tables(
        database_name,
        location_prefix,
        output_location=output_location,
        replace=replace_tables,
    )


@athena.command("query")
@click.option(
    "--database-name", "-d", default="swh", help="Name of the database to query"
)
@click.option(
    "-o", "--output-location", help="S3 prefix where results should be stored"
)
@click.argument("query_file", type=click.File("r"), default=sys.stdin)
def athena_query(
    database_name,
    query_file,
    output_location=None,
):
    """Query the AWS Athena database with a given command"""
    from .athena import run_query_get_results

    print(
        run_query_get_results(
            database_name,
            query_file.read(),
            output_location=output_location,
        ),
        end="",
    )  # CSV already ends with \n


@athena.command("gensubdataset")
@click.option("--database", "-d", default="swh", help="Name of the base database")
@click.option(
    "--subdataset-database",
    required=True,
    help="Name of the subdataset database to create",
)
@click.option(
    "--subdataset-location",
    required=True,
    help="S3 prefix where the subdataset should be stored",
)
@click.option(
    "--swhids",
    required=True,
    help="File containing the list of SWHIDs to include in the subdataset",
)
def athena_gensubdataset(database, subdataset_database, subdataset_location, swhids):
    """
    Generate a subdataset with Athena, from an existing database and a list
    of SWHIDs. Athena will generate a new dataset with the same tables as in
    the base dataset, but only containing the objects present in the SWHID
    list.
    """
    from .athena import generate_subdataset

    generate_subdataset(
        database,
        subdataset_database,
        subdataset_location,
        swhids,
        os.path.join(subdataset_location, "queries"),
    )



# Copyright (C) 2025 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import base64
import csv
import os
from pathlib import Path
import subprocess
import tempfile

import pyorc
from tqdm import tqdm


def process_fullnames(fullnames_orc: Path, dedup_dir: Path) -> None:
    with tempfile.NamedTemporaryFile(suffix=".csv") as result_file:
        entries = list(dedup_dir.iterdir())
        if entries:
            env = {**os.environ, "LC_ALL": "C", "LC_COLLATE": "C", "LANG": "C"}
            # fmt: off
            subprocess.run(
                [
                    "sort",
                    "-t", ",",
                    "-k", "2",
                    "-u",
                    "-S", "100M",
                    "-m",
                    *entries,
                    "-o", result_file.name,
                ],
                env=env,
            )
            # fmt: on

        with open(fullnames_orc, "wb") as output:
            with open(result_file.name, "r") as input:
                reader = csv.reader(input)
                with pyorc.Writer(
                    output,
                    pyorc.Struct(
                        fullname=pyorc.Binary(), sha256_fullname=pyorc.Binary()
                    ),
                    bloom_filter_columns=[0, 1],
                ) as writer:
                    for row in tqdm(
                        reader, desc="Writing persons' fullnames to ORC file"
                    ):
                        if row == ("",):
                            continue
                        fullname, sha256_fullname = row
                        writer.write(
                            (
                                base64.b64decode(fullname),
                                base64.b64decode(sha256_fullname),
                            )
                        )



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import functools
import gc
import multiprocessing
from multiprocessing.managers import ListProxy
from typing import Sequence, Tuple

import pytest

from swh.export.exporter import Exporter
from swh.export.journalprocessor import ParallelJournalProcessor
from swh.journal.writer import get_journal_writer
from swh.model import model
from swh.model.tests import swh_model_data


@pytest.fixture
def journal_client_config(
    kafka_server: str, kafka_prefix: str, kafka_consumer_group: str
):
    return dict(
        brokers=kafka_server,
        group_id=kafka_consumer_group,
        prefix=kafka_prefix,
    )


@pytest.fixture
def journal_writer(kafka_server: str, kafka_prefix: str):
    return get_journal_writer(
        cls="kafka",
        brokers=[kafka_server],
        client_id="kafka_writer",
        prefix=kafka_prefix,
        anonymize=False,
    )


def disable_gc(f):
    """Decorator for test functions; prevents segfaults in confluent-kafka.
    See https://github.com/confluentinc/confluent-kafka-python/issues/1761"""

    @functools.wraps(f)
    def newf(*args, **kwargs):
        gc.disable()
        try:
            return f(*args, **kwargs)
        finally:
            gc.enable()

    return newf


class ListExporter(Exporter):
    def __init__(self, objects: ListProxy, *args, **kwargs):
        self._objects = objects
        super().__init__(*args, **kwargs)

    def process_object(
        self, object_type: model.ModelObjectType, obj: model.BaseModel
    ) -> None:
        self._objects.append((object_type, obj))


def assert_exported_objects(
    exported_objects: Sequence[Tuple[str, model.BaseModel]],
    expected_objects: Sequence[model.BaseModel],
) -> None:
    assert set(obj[1] for obj in exported_objects) == set(expected_objects)


def test_parallel_journal_processor(
    journal_client_config, journal_writer, tmp_path
) -> None:
    journal_writer.write_additions("revision", swh_model_data.REVISIONS)

    with multiprocessing.Manager() as manager:
        objects = manager.list()
        (tmp_path / "node_sets" / "revision").mkdir(parents=True)
        export_path = tmp_path / "export"
        sensitive_export_path = tmp_path / "sensitive" / "export"
        dup_dir = tmp_path / "duplicates"
        dup_dir.mkdir(parents=True, exist_ok=True)
        dedup_dir = tmp_path / "deduplicated"
        dedup_dir.mkdir(parents=True, exist_ok=True)
        config = {"journal": journal_client_config}
        processor = ParallelJournalProcessor(
            config=config,
            masked_swhids=set(),
            exporter_factories=[
                functools.partial(
                    ListExporter,
                    objects,
                    config=config,
                    export_path=export_path,
                    sensitive_export_path=sensitive_export_path,
                )
            ],
            export_id="test_parallel_journal_processor",
            obj_type="revision",
            node_sets_path=tmp_path / "node_sets",
            dup_dir=dup_dir,
            dedup_dir=dedup_dir,
        )

        processor.run()

        assert_exported_objects(objects, swh_model_data.REVISIONS)


def test_parallel_journal_processor_origin(
    journal_client_config, journal_writer, tmp_path
) -> None:
    journal_writer.write_additions("origin", swh_model_data.ORIGINS)

    with multiprocessing.Manager() as manager:
        objects = manager.list()
        (tmp_path / "node_sets" / "origin").mkdir(parents=True)
        export_path = tmp_path / "export"
        sensitive_export_path = tmp_path / "sensitive" / "export"
        dup_dir = tmp_path / "duplicates"
        dup_dir.mkdir(parents=True, exist_ok=True)
        dedup_dir = tmp_path / "deduplicated"
        dedup_dir.mkdir(parents=True, exist_ok=True)
        config = {"journal": journal_client_config}
        processor = ParallelJournalProcessor(
            config=config,
            masked_swhids=set(),
            exporter_factories=[
                functools.partial(
                    ListExporter,
                    objects,
                    config=config,
                    export_path=export_path,
                    sensitive_export_path=sensitive_export_path,
                )
            ],
            export_id="test_parallel_journal_processor",
            obj_type="origin",
            node_sets_path=tmp_path / "node_sets",
            dup_dir=dup_dir,
            dedup_dir=dedup_dir,
        )

        processor.run()

        assert_exported_objects(objects, swh_model_data.ORIGINS)


@disable_gc
def test_parallel_journal_processor_origin_visit_status(
    journal_client_config, journal_writer, tmp_path
) -> None:
    journal_writer.write_additions(
        "origin_visit_status", swh_model_data.ORIGIN_VISIT_STATUSES
    )

    with multiprocessing.Manager() as manager:
        objects = manager.list()
        (tmp_path / "node_sets" / "origin_visit_status").mkdir(parents=True)
        export_path = tmp_path / "export"
        sensitive_export_path = tmp_path / "sensitive" / "export"
        dup_dir = tmp_path / "duplicates"
        dup_dir.mkdir(parents=True, exist_ok=True)
        dedup_dir = tmp_path / "deduplicated"
        dedup_dir.mkdir(parents=True, exist_ok=True)
        config = {"journal": journal_client_config}
        processor = ParallelJournalProcessor(
            config=config,
            masked_swhids=set(),
            exporter_factories=[
                functools.partial(
                    ListExporter,
                    objects,
                    config=config,
                    export_path=export_path,
                    sensitive_export_path=sensitive_export_path,
                )
            ],
            export_id="test_parallel_journal_processor",
            obj_type="origin_visit_status",
            node_sets_path=tmp_path / "node_sets",
            dup_dir=dup_dir,
            dedup_dir=dedup_dir,
        )

        processor.run()

        assert_exported_objects(objects, swh_model_data.ORIGIN_VISIT_STATUSES)


@disable_gc
def test_parallel_journal_processor_offsets(
    journal_client_config, journal_writer, tmp_path
) -> None:
    """Checks the exporter stops at the offsets computed at the beginning of the export"""
    journal_writer.write_additions("revision", swh_model_data.REVISIONS[0:2])

    with multiprocessing.Manager() as manager:
        objects = manager.list()
        (tmp_path / "node_sets" / "revision").mkdir(parents=True)
        export_path = tmp_path / "export"
        sensitive_export_path = tmp_path / "sensitive" / "export"
        dup_dir = tmp_path / "duplicates"
        dup_dir.mkdir(parents=True, exist_ok=True)
        dedup_dir = tmp_path / "deduplicated"
        dedup_dir.mkdir(parents=True, exist_ok=True)
        config = {"journal": journal_client_config}
        processor = ParallelJournalProcessor(
            config=config,
            masked_swhids=set(),
            exporter_factories=[
                functools.partial(
                    ListExporter,
                    objects,
                    config=config,
                    export_path=export_path,
                    sensitive_export_path=sensitive_export_path,
                )
            ],
            export_id="test_parallel_journal_processor",
            obj_type="revision",
            node_sets_path=tmp_path / "node_sets",
            dup_dir=dup_dir,
            dedup_dir=dedup_dir,
        )

        processor.get_offsets()  # fills the processor.offsets cache

        processor.run()

        assert_exported_objects(objects, swh_model_data.REVISIONS[0:2])


@disable_gc
def test_parallel_journal_processor_masked(
    journal_client_config, journal_writer, tmp_path
) -> None:
    journal_writer.write_additions("revision", swh_model_data.REVISIONS)

    masked_swhids = {swh_model_data.REVISIONS[2].swhid().to_extended()}

    with multiprocessing.Manager() as manager:
        objects = manager.list()
        (tmp_path / "node_sets" / "revision").mkdir(parents=True)
        export_path = tmp_path / "export"
        sensitive_export_path = tmp_path / "sensitive" / "export"
        dup_dir = tmp_path / "duplicates"
        dup_dir.mkdir(parents=True, exist_ok=True)
        dedup_dir = tmp_path / "deduplicated"
        dedup_dir.mkdir(parents=True, exist_ok=True)
        config = {"journal": journal_client_config}
        processor = ParallelJournalProcessor(
            config=config,
            masked_swhids=masked_swhids,
            exporter_factories=[
                functools.partial(
                    ListExporter,
                    objects,
                    config=config,
                    export_path=export_path,
                    sensitive_export_path=sensitive_export_path,
                )
            ],
            export_id="test_parallel_journal_processor",
            obj_type="revision",
            node_sets_path=tmp_path / "node_sets",
            dup_dir=dup_dir,
            dedup_dir=dedup_dir,
        )

        processor.run()

        assert_exported_objects(
            objects, swh_model_data.REVISIONS[0:2] + swh_model_data.REVISIONS[3:]
        )


@disable_gc
def test_parallel_journal_processor_masked_origin(
    journal_client_config, journal_writer, tmp_path
) -> None:
    journal_writer.write_additions("origin", swh_model_data.ORIGINS)

    masked_swhids = {swh_model_data.ORIGINS[1].swhid()}

    with multiprocessing.Manager() as manager:
        objects = manager.list()
        (tmp_path / "node_sets" / "origin").mkdir(parents=True)
        export_path = tmp_path / "export"
        sensitive_export_path = tmp_path / "sensitive" / "export"
        dup_dir = tmp_path / "duplicates"
        dup_dir.mkdir(parents=True, exist_ok=True)
        dedup_dir = tmp_path / "deduplicated"
        dedup_dir.mkdir(parents=True, exist_ok=True)
        config = {"journal": journal_client_config}
        processor = ParallelJournalProcessor(
            config=config,
            masked_swhids=masked_swhids,
            exporter_factories=[
                functools.partial(
                    ListExporter,
                    objects,
                    config=config,
                    export_path=export_path,
                    sensitive_export_path=sensitive_export_path,
                )
            ],
            export_id="test_parallel_journal_processor",
            obj_type="origin",
            node_sets_path=tmp_path / "node_sets",
            dup_dir=dup_dir,
            dedup_dir=dedup_dir,
        )

        processor.run()

        assert_exported_objects(
            objects, swh_model_data.ORIGINS[0:1] + swh_model_data.ORIGINS[2:]
        )


@disable_gc
def test_parallel_journal_processor_masked_origin_visit_statuses(
    journal_client_config, journal_writer, tmp_path
) -> None:
    journal_writer.write_additions(
        "origin_visit_status", swh_model_data.ORIGIN_VISIT_STATUSES
    )

    masked_origin = model.Origin(url=swh_model_data.ORIGIN_VISIT_STATUSES[1].origin)
    masked_swhids = {masked_origin.swhid()}

    with multiprocessing.Manager() as manager:
        objects = manager.list()
        (tmp_path / "node_sets" / "origin_visit_status").mkdir(parents=True)
        export_path = tmp_path / "export"
        sensitive_export_path = tmp_path / "sensitive" / "export"
        dup_dir = tmp_path / "duplicates"
        dup_dir.mkdir(parents=True, exist_ok=True)
        dedup_dir = tmp_path / "deduplicated"
        dedup_dir.mkdir(parents=True, exist_ok=True)
        config = {"journal": journal_client_config}
        processor = ParallelJournalProcessor(
            config=config,
            masked_swhids=masked_swhids,
            exporter_factories=[
                functools.partial(
                    ListExporter,
                    objects,
                    config=config,
                    export_path=export_path,
                    sensitive_export_path=sensitive_export_path,
                )
            ],
            export_id="test_parallel_journal_processor",
            obj_type="origin_visit_status",
            node_sets_path=tmp_path / "node_sets",
            dup_dir=dup_dir,
            dedup_dir=dedup_dir,
        )

        processor.run()

        assert_exported_objects(
            objects,
            [
                ovs
                for ovs in swh_model_data.ORIGIN_VISIT_STATUSES
                if ovs.origin != masked_origin.url
            ],
        )



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import contextlib
import pathlib
from types import TracebackType
from typing import Any, Dict, Optional, Type
import uuid

from swh.model.model import BaseModel, ModelObjectType


class Exporter:
    """
    Base class for all the exporters.

    Each export can have multiple exporters, so we can read the journal a single
    time, then export the objects we read in different formats without having to
    re-read them every time.

    Override this class with the behavior for an export in a specific export
    format. You have to overwrite process_object() to make it write to the
    appropriate export files.

    You can also put setup and teardown logic in __enter__ and __exit__, and it
    will be called automatically.
    """

    def __init__(
        self,
        config: Dict[str, Any],
        export_path,
        sensitive_export_path,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.config: Dict[str, Any] = config
        self.export_path = pathlib.Path(export_path)
        if sensitive_export_path is not None:
            self.sensitive_export_path = pathlib.Path(sensitive_export_path)
        else:
            self.sensitive_export_path = (
                self.export_path.parent / f"{self.export_path.name}-sensitive"
            )
        self.exit_stack = contextlib.ExitStack()

    def __enter__(self) -> "Exporter":
        self.export_path.mkdir(exist_ok=True, parents=True)
        self.sensitive_export_path.mkdir(exist_ok=True, parents=True)
        self.exit_stack.__enter__()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        return self.exit_stack.__exit__(exc_type, exc_value, traceback)

    def process_object(self, object_type: ModelObjectType, obj: BaseModel) -> None:
        """
        Process a SWH object to export.

        Override this with your custom exporter.
        """
        raise NotImplementedError

    def get_unique_file_id(self) -> str:
        """
        Return a unique random file id for the current process.

        If config['test_unique_file_id'] is set, it will be used instead.
        """
        return str(self.config.get("test_unique_file_id", uuid.uuid4()))


class ExporterDispatch(Exporter):
    """
    Like Exporter, but dispatches each object type to a different function
    (e.g you can override `process_origin(self, object)` to process origins.)
    """

    def process_object(self, object_type: ModelObjectType, obj: BaseModel) -> None:
        method_name = "process_" + object_type.name.lower()
        if hasattr(self, method_name):
            getattr(self, method_name)(obj)



# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Dict, List, Union

# fmt: off
MAIN_TABLES = {
    "origin": [
        ("id", "string"),
        ("url", "string"),
    ],
    "origin_visit": [
        ("origin", "string"),
        ("visit", "bigint"),
        ("date", "timestamp"),
        ("type", "string"),
    ],
    "origin_visit_status": [
        ("origin", "string"),
        ("visit", "bigint"),
        ("date", "timestamp"),
        ("status", "string"),
        ("snapshot", "string"),
        ("type", "string"),
    ],
    "snapshot": [
        ("id", "string"),
    ],
    # snapshot_branches is in RELATED_TABLES
    "release": [
        ("id", "string"),
        ("name", "binary"),
        ("message", "binary"),
        ("target", "string"),
        ("target_type", "string"),
        ("author", "binary"),
        ("date", "timestamp"),
        ("date_offset", "smallint"),
        ("date_raw_offset_bytes", "binary"),
        ("raw_manifest", "binary"),
    ],
    "revision": [
        ("id", "string"),
        ("message", "binary"),
        ("author", "binary"),
        ("date", "timestamp"),
        ("date_offset", "smallint"),
        ("date_raw_offset_bytes", "binary"),
        ("committer", "binary"),
        ("committer_date", "timestamp"),
        ("committer_offset", "smallint"),  # called committer_date_offset in swh-storage
        ("committer_date_raw_offset_bytes", "binary"),
        ("directory", "string"),
        ("type", "string"),
        ("raw_manifest", "binary"),
    ],
    # revision_history is in RELATED_TABLES
    # revision_extra_headers is in RELATED_TABLES
    "directory": [
        ("id", "string"),
        ("raw_manifest", "binary"),
    ],
    # direcory_entry is in RELATED_TABLES
    "content": [
        ("sha1", "string"),
        ("sha1_git", "string"),
        ("sha256", "string"),
        ("blake2s256", "string"),
        ("length", "bigint"),
        ("status", "string"),
        ("data", "binary")
    ],
    "skipped_content": [
        ("sha1", "string"),
        ("sha1_git", "string"),
        ("sha256", "string"),
        ("blake2s256", "string"),
        ("length", "bigint"),
        ("status", "string"),
        ("reason", "string"),
    ],
}

RELATION_TABLES = {
    "snapshot_branch": [
        ("snapshot_id", "string"),
        ("name", "binary"),
        ("target", "string"),
        ("target_type", "string"),
    ],
    "revision_history": [
        ("id", "string"),
        ("parent_id", "string"),
        ("parent_rank", "int"),
    ],
    "revision_extra_headers": [
        ("id", "string"),
        ("key", "binary"),
        ("value", "binary"),
    ],
    "directory_entry": [
        ("directory_id", "string"),
        ("name", "binary"),
        ("type", "string"),
        ("target", "string"),
        ("perms", "int"),
    ],
}

TABLES = {**MAIN_TABLES, **RELATION_TABLES}

BLOOM_FILTER_COLUMNS: Dict[str, List[Union[str, int]]] = {
    "origin": [
        "url",  # allows checking if an origin is in the dataset
    ],
    "origin_visit": [
        "origin",  # allows listing visits of an origin
    ],
    "origin_visit_status": [
        "origin",  # allows listing visit statuses of an origin
    ],
    "snapshot": [
        "id",  # allows checking a snapshot is in the dataset
    ],
    "release": [
        "id",
        "author",
        "target",  # allows reverse traversal
    ],
    "revision": [
        "id",
        "author",
        "committer",
        "directory",  # allows reverse traversal
    ],
    "directory": [
        "id",  # allows checking a directory is in the dataset
    ],
    "content": [
        "sha1",
        "sha1_git",
        "sha256",
        # not including blake2s256, it's unlikely to be useful
    ],
    "skipped_content": [
        "sha1",
        "sha1_git",
        "sha256",
        # not including blake2s256, it's unlikely to be useful
    ],
    "snapshot_branch": [
        "snapshot_id",  # allows finding the stripe containing a snapshot's branches
        "target",  # allows reverse traversal
    ],
    "revision_history": [
        "id",  # allows finding the stripe containing a revision's parents
        "parent_id",  # allows reverse traversal
    ],
    "revision_extra_headers": [],
    "directory_entry": [
        "directory_id",  # allows finding the stripe containing a revision's parents
        "target",  # allows reverse traversal
        # not including name, it is often filtered with wildcards and/or case
        # normalization, so Bloom Filters cannot be used.
    ],
}
"""
Columns where we include Bloom filters.

They allow looking for high cardinality values without decompressing most stripes
not containing any (equality) match.
"""

# fmt: on
